# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
"""
Typing information for generated version module.
"""

pyosmium_release: str
libosmium_version: str
protozero_version: str
pybind11_version: str
