# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import datetime

def newest_change_from_file(filename: str) -> datetime.datetime:
    """ Find the data of the most recent change in a file.
    """
